package pins25.phase;

import pins25.common.*;

import java.util.Objects;

/**
 * Sintaksni analizator.
 */
public class SynAn implements AutoCloseable {

    /**
     * Leksikalni analizator.
     */
    private final LexAn lexAn;

    /**
     * Ustvari nov sintaksni analizator.
     *
     * @param srcFileName Ime izvorne datoteke.
     */
    public SynAn(final String srcFileName) {
        this.lexAn = new LexAn(srcFileName);
    }

    @Override
    public void close() {
        lexAn.close();
    }

    /**
     * Prevzame leksikalni analizator od leksikalnega analizatorja in preveri, ali
     * je prave vrste.
     *
     * @param symbol Pričakovana vrsta leksikalnega simbola.
     * @return Prevzeti leksikalni simbol.
     */
    private Token check(Token.Symbol symbol) {
        final Token token = lexAn.takeToken();
        if (token.symbol() != symbol)
            throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected " + symbol + ".");
        return token;
    }

    /**
     * Opravi sintaksno analizo.
     */
    public void parse() {
        parseProgram();

        if (lexAn.peekToken().symbol() != Token.Symbol.EOF)
            Report.warning(lexAn.peekToken(),
                    "Unexpected text '" + lexAn.peekToken().lexeme() + "...' at the end of the program.");
    }

    private void parseProgram() {
        parseDefinition();
        parseRestDefinitions();
    }

    private void parseRestDefinitions() {
        if (lexAn.peekToken().symbol() != Token.Symbol.EOF) {
            parseDefinition();
            parseRestDefinitions();
        }
    }

    private void parseDefinition() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case Token.Symbol.FUN:
                check(Token.Symbol.FUN);
                check(Token.Symbol.IDENTIFIER);
                check(Token.Symbol.LPAREN);
                parseParameters();
                check(Token.Symbol.RPAREN);
                parseFunctionAssignment();
                break;

            case Token.Symbol.VAR:
                check(Token.Symbol.VAR);
                check(Token.Symbol.IDENTIFIER);
                check(Token.Symbol.ASSIGN);
                parseInitializers();
                break;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected FUN or VAR.");
        }
    }

    private void parseFunctionAssignment() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case Token.Symbol.FUN:
            case Token.Symbol.VAR:
            case Token.Symbol.IN:
            case Token.Symbol.EOF:
                break;

            case Token.Symbol.ASSIGN:
                check(Token.Symbol.ASSIGN);
                parseStatements();
                break;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected ASSIGN.");
        }
    }

    private void parseParameters() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case Token.Symbol.IDENTIFIER:
                check(Token.Symbol.IDENTIFIER);
                parseRestParameters();
                break;

            case Token.Symbol.RPAREN:
                break;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected IDENTIFIER or RPAREN.");
        }
    }

    private void parseRestParameters() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case Token.Symbol.RPAREN:
                break;

            case Token.Symbol.COMMA:
                check(Token.Symbol.COMMA);
                check(Token.Symbol.IDENTIFIER);
                parseRestParameters();
                break;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected ')' or ','.");
        }
    }

    private void parseStatements() {
        parseStatement();
        parseRestStatements();
    }

    private void parseStatement() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case Token.Symbol.IDENTIFIER:
            case Token.Symbol.LPAREN:
            case Token.Symbol.ADD:
            case Token.Symbol.SUB:
            case Token.Symbol.NOT:
            case Token.Symbol.PTR:
            case Token.Symbol.INTCONST:
            case Token.Symbol.CHARCONST:
            case Token.Symbol.STRINGCONST:
                parseExpression();
                parseExpressionAssign();
                break;

            case Token.Symbol.IF:
                check(Token.Symbol.IF);
                parseExpression();
                check(Token.Symbol.THEN);
                parseStatements();
                parseElseStatement();
                check(Token.Symbol.END);
                break;

            case Token.Symbol.WHILE:
                check(Token.Symbol.WHILE);
                parseExpression();
                check(Token.Symbol.DO);
                parseStatements();
                check(Token.Symbol.END);
                break;

            case Token.Symbol.LET:
                check(Token.Symbol.LET);
                parseDefinition();
                parseRestStatementDefinitions();
                check(Token.Symbol.IN);
                parseStatements();
                check(Token.Symbol.END);
                break;
        }
    }

    private void parseRestStatements() {
        if (lexAn.peekToken().symbol() == Token.Symbol.COMMA) {
            check(Token.Symbol.COMMA);
            parseStatements();
        }
    }

    private void parseRestStatementDefinitions() {

    }

    private void parseInitializers() {

    }

    private void parseExpression() {
    }

    private void parseExpressionAssign() {
    }

    private void parseElseStatement() {
    }

    // --- ZAGON ---

    /**
     * Zagon sintaksnega analizatorja kot samostojnega programa.
     *
     * @param cmdLineArgs Argumenti v ukazni vrstici.
     */
    public static void main(final String[] cmdLineArgs) {
        System.out.println("This is PINS'25 compiler (syntax analysis):");

        try {
            if (cmdLineArgs.length == 0)
                throw new Report.Error("No source file specified in the command line.");
            if (cmdLineArgs.length > 1)
                Report.warning("Unused arguments in the command line.");

            try (SynAn synAn = new SynAn(cmdLineArgs[0])) {
                synAn.parse();
            }

            // Upajmo, da kdaj pridemo to te tocke.
            // A zavedajmo se sledecega:
            // 1. Prevod je zaradi napak v programu lahko napacen :-o
            // 2. Izvorni program se zdalec ni tisto, kar je programer hotel, da bi bil ;-)
            Report.info("Done.");
        } catch (Report.Error error) {
            // Izpis opisa napake.
            System.err.println(error.getMessage());
            System.exit(1);
        }
    }

}
